import pytest
import sdmx
from sdmx.format import Version
from sdmx.model.common import Code, Codelist

import genno.operator
from genno import Computer
from genno.compat.sdmx import operator
from genno.testing import add_test_data


def test_codelist_to_groups() -> None:
    c = Computer()
    _, t_foo, t_bar, __ = add_test_data(c)

    cl: Codelist = Codelist(id="t")
    cl.append(Code(id="foo", child=[Code(id=t) for t in t_foo]))
    cl.append(Code(id="bar", child=[Code(id=t) for t in t_bar]))

    # Operator runs
    for result0 in (
        operator.codelist_to_groups(cl),
        operator.codelist_to_groups(iter(cl), dim="t"),
    ):
        # Result has the expected contents
        assert {"t"} == set(result0.keys())
        result_t = result0["t"]
        assert {"foo", "bar"} == set(result_t.keys())
        assert set(t_foo) == set(result_t["foo"])
        assert set(t_bar) == set(result_t["bar"])

    with pytest.raises(ValueError, match="Must provide a dimension"):
        operator.codelist_to_groups(iter(cl))

    # Output is usable in Computer() with aggregate
    c.require_compat("genno.compat.sdmx")
    c.add("t::codes", cl)
    c.add("t::groups", "codelist_to_groups", "t::codes")
    key = c.add("x::agg", "aggregate", "x:t-y", "t::groups", False)

    result1 = c.get(key)

    # Quantity was aggregated per `cl`
    assert {"foo", "bar"} == set(result1.coords["t"].data)


@pytest.fixture(scope="session")
def dsd(test_data_path):
    # Read the data structure definition
    yield sdmx.read_sdmx(test_data_path.joinpath("22_289-structure.xml")).structure[
        "DCIS_POPRES1"
    ]


@pytest.fixture(scope="session")
def dm(test_data_path, dsd):
    # Read the data message
    yield sdmx.read_sdmx(test_data_path.joinpath("22_289.xml"), structure=dsd)


def test_dataset_to_quantity(dsd, dm) -> None:
    # Select the data set
    ds = dm.data[0]

    # Operator runs
    result = operator.dataset_to_quantity(ds)

    # Dimensions of the quantity match the dimensions of the data frame
    assert set(d.id for d in dsd.dimensions.components) == set(result.dims)

    # Attributes contain information on the data set and its structure
    assert (
        "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructureDefinition="
        "IT1:DCIS_POPRES1(1.0)" == result.attrs["structure_urn"]
    )

    # All observations are converted
    assert len(ds.obs) == result.size


VERSION = (None, Version["2.1"], Version["3.0"], "2.1", "3.0")


@pytest.mark.parametrize("observation_dimension", (None, "TIME_PERIOD"))
@pytest.mark.parametrize("version", VERSION)
@pytest.mark.parametrize("with_attrs", (True, False))
def test_quantity_to_dataset(
    dsd, dm, observation_dimension, version, with_attrs
) -> None:
    ds = dm.data[0]
    qty = operator.dataset_to_quantity(ds)

    if not with_attrs:
        qty.attrs.pop("structure_urn")

    result = operator.quantity_to_dataset(
        qty, structure=dsd, observation_dimension=observation_dimension, version=version
    )

    # All observations are converted
    assert len(ds.obs) == len(result.obs)

    # Dataset is associated with its DSD
    assert dsd is result.structured_by


@pytest.mark.parametrize("observation_dimension", (None, "TIME_PERIOD"))
@pytest.mark.parametrize("version", VERSION)
def test_quantity_to_message(dsd, dm, observation_dimension, version) -> None:
    ds = dm.data[0]
    qty = operator.dataset_to_quantity(ds)

    header = dm.header

    result = operator.quantity_to_message(
        qty,
        structure=dsd,
        observation_dimension=observation_dimension,
        version=version,
        header=header,
    )

    # Currently False because OBS_STATUS attributes are not preserved
    with pytest.raises(AssertionError):
        # Resulting message compares equal to the original ("round trip")
        assert dm.compare(result)


@pytest.mark.parametrize("observation_dimension", (None, "TIME_PERIOD"))
@pytest.mark.parametrize(
    "version",
    (
        None,
        Version["2.1"],
        pytest.param(
            Version["3.0"],
            marks=pytest.mark.xfail(
                raises=NotImplementedError, reason="Not implemented in sdmx1"
            ),
        ),
        "2.1",
        pytest.param(
            "3.0",
            marks=pytest.mark.xfail(
                raises=NotImplementedError, reason="Not implemented in sdmx1"
            ),
        ),
    ),
)
def test_write_report(tmp_path, dsd, dm, observation_dimension, version) -> None:
    ds = dm.data[0]
    qty = operator.dataset_to_quantity(ds)

    header = dm.header

    # Quantity can be converted to DataMessage
    obj = operator.quantity_to_message(
        qty,
        structure=dsd,
        observation_dimension=observation_dimension,
        version=version,
        header=header,
    )

    path = tmp_path.joinpath("foo.xml")

    # File can be written
    genno.operator.write_report(obj, path)

    # File contains the SDMX namespaces
    # NB Could do more extensive tests of the file contents here
    assert "http://www.sdmx.org/resources/sdmxml/schemas/v2_1" in path.read_text()



from dask.core import literal, quote

from .config import configure
from .core.computer import Computer
from .core.exceptions import ComputationError, KeyExistsError, MissingKeyError
from .core.key import Key, KeySeq
from .core.operator import Operator
from .core.quantity import Quantity, assert_quantity, get_class, set_class

__all__ = [
    "ComputationError",
    "Computer",
    "Key",
    "KeySeq",
    "KeyExistsError",
    "MissingKeyError",
    "Operator",
    "Quantity",
    "assert_quantity",
    "configure",
    "get_class",
    "literal",
    "quote",
    "set_class",
]



# This module should only be imported during type checking
from typing import TYPE_CHECKING  # pragma: no cover

if TYPE_CHECKING:
    from pint import Unit
    from xarray.core.types import Dims, InterpOptions, ScalarOrArray

    from .core.quantity import AnyQuantity

    __all__ = [
        "AnyQuantity",
        "Dims",
        "InterpOptions",
        "ScalarOrArray",
        "Unit",
    ]



import logging
from functools import partial
from inspect import Parameter, signature
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    Union,
)

import numpy as np
import pandas as pd
from dask.core import literal

from .compat.xarray import is_scalar
from .core.key import Key

if TYPE_CHECKING:
    import pint

log = logging.getLogger(__name__)


#: Replacements to apply to Quantity units before parsing by
#: :doc:`pint <pint:index>`. Mapping from original unit -> preferred unit.
#:
#: The default values include:
#:
#: - The '%' symbol cannot be supported by pint, because it is a Python operator; it is
#:   replaced with “percent”.
#:
#: Additional values can be added with :func:`.configure`; see :ref:`config-units`.
REPLACE_UNITS = {
    "%": "percent",
}

# For use in type hints
UnitLike = Union[str, "pint.Unit", "pint.Quantity"]


def clean_units(input_string):
    """Tolerate messy strings for units.

    - Dimensions enclosed in “[]” have these characters stripped.
    - Replacements from :data:`.REPLACE_UNITS` are applied.
    """
    input_string = input_string.strip("[]")
    for old, new in REPLACE_UNITS.items():
        input_string = input_string.replace(old, new)
    return input_string


def collect_units(*args):
    """Return the "_unit" attributes of the `args`."""
    import pint

    registry = pint.get_application_registry()

    result = []
    for arg in args:
        try:
            unit = arg.attrs.get("_unit")
        except AttributeError:
            if is_scalar(arg):
                result.append(registry.dimensionless)
            else:
                raise  # pragma: no cover
        else:
            if unit is None:
                log.debug(
                    f"{arg.__class__.__name__} '{arg.name or '(no name)'}' {arg.dims!r}"
                    " lacks units; assume dimensionless"
                )
                unit = registry.dimensionless

            # Convert a possible string or other expression to a pint.Unit object
            arg.units = registry.Unit(unit)

            result.append(arg.units)

    return tuple(result)


def filter_concat_args(args):
    """Filter out str and Key from *args*.

    A warning is logged for each element removed.
    """
    for arg in args:
        if isinstance(arg, (str, Key)):
            log.warning(f"concat() argument {repr(arg)} missing; will be omitted")
            continue
        yield arg


def _invalid(unit: str, exc: Exception) -> Exception:
    """Helper method to return an intelligible exception from :func:`parse_units`."""
    chars = "".join(filter("-?$".__contains__, unit))
    msg = f"unit {unit!r} cannot be parsed; contains invalid character(s) {chars!r}"
    # Use the original class of `exc`, mapped in some cases
    cls_map: Mapping[type[Exception], type[Exception]] = {TypeError: ValueError}
    return_cls = cls_map.get(type(exc), type(exc))
    return return_cls(msg)


def parse_units(data: Iterable, registry=None) -> "pint.Unit":
    """Return a :class:`pint.Unit` for an iterable of strings.

    Valid unit expressions not already present in the `registry` are defined, e.g.:

    .. code-block:: python

       u = parse_units(["foo/bar", "foo/bar"], reg)

    …results in the addition of unit definitions equivalent to:

    .. code-block:: python

       reg.define("foo = [foo]")
       reg.define("bar = [bar]")
       u = reg.foo / reg.bar

    Raises
    ------
    ValueError
        if `data` contains more than 1 unit expression, or the unit expression contains
        characters not parseable by :mod:`pint`, e.g. ``-?$``.
    """
    import pint

    from .compat.pint import PintError

    registry = registry or pint.get_application_registry()

    # Ensure a type that is accepted by pd.unique()
    if isinstance(data, str):
        data = np.array([data])
    elif not isinstance(data, (np.ndarray, pd.Index, pd.Series)):
        data = np.array(data)

    unit = pd.unique(data)

    if len(unit) > 1:
        raise ValueError(f"mixed units {list(unit)}")

    try:
        unit = clean_units(unit[0])
    except IndexError:
        # `units_series` is length 0 → no data → dimensionless
        unit = registry.dimensionless

    # Parse units
    try:
        return registry.Unit(unit)
    except pint.UndefinedUnitError:
        try:
            # Unit(s) do not exist; define them in the UnitRegistry
            # TODO add global configuration to disable this feature.
            # Split possible compound units
            for part in unit.split("/"):
                try:
                    registry.Unit(part)
                except pint.UndefinedUnitError:
                    # Part was unparseable; define it
                    definition = f"{part} = [{part}]"
                    log.info(f"Add unit definition: {definition}")

                    # This line will fail silently for parts like 'G$' containing
                    # characters like '$' that are discarded by pint
                    registry.define(definition)

            # Try to parse again
            return registry.Unit(unit)
        except PintError as e:
            # registry.define() failed somehow
            raise _invalid(unit, e)
    except (AttributeError, TypeError) + PintError as e:  # type: ignore [misc]
        # Unit contains a character like '-' that throws off pint
        # NB this 'except' clause must be *after* UndefinedUnitError, since that is a
        #    subclass of AttributeError.
        raise _invalid(unit, e)


_pars_cache: dict[tuple[Callable, int, tuple], Mapping] = {}


def free_parameters(func: Callable) -> Mapping:
    """Retrieve information on the free parameters of `func`.

    Identical to :py:`inspect.signature(func).parameters`; that is, to
    :attr:`inspect.Signature.parameters`. :py:`free_parameters` also:

    - Handles functions that have been :func:`functools.partial`'d, returning only the
      parameters that have *not* already been assigned a value by the
      :func:`~functools.partial` call—the “free” parameters.
    - Caches return values for better performance.
    """

    # Form a cache key; possibly unwrap information from a partialled function
    key = (
        getattr(func, "func", func),  # The base callable or function
        len(getattr(func, "args", [])),  # Number of positional args partialled
        tuple(sorted(getattr(func, "keywords", {}))),  # Names of partialled kw args
    )

    try:
        return _pars_cache[key]
    except KeyError:
        try:
            result: Mapping = signature(func).parameters
        except ValueError:
            # signature() raises for operator.itemgetter(…), built-ins, and similar
            if not callable(func):  # pragma: no cover
                raise TypeError(type(func))
            result = {}

        return _pars_cache.setdefault(key, result)


def partial_split(func: Callable, kwargs: Mapping) -> tuple[Callable, MutableMapping]:
    """Forgiving version of :func:`functools.partial`.

    Returns a :ref:`partial object <python:partial-objects>` and leftover keyword
    arguments that are not applicable to `func`.
    """
    # Retrieve information on the free parameters of `func`
    pars = free_parameters(func)

    func_args, extra = {}, {}
    for name, value in kwargs.items():
        if name in pars and pars[name].kind in (
            Parameter.POSITIONAL_OR_KEYWORD,
            Parameter.KEYWORD_ONLY,
        ):
            # A keyword argument that can be passed to `func`
            func_args[name] = value
        else:
            extra[name] = value

    if func_args:
        return partial(func, **func_args), extra
    else:
        return func, extra  # Nothing to partial; return `func` as-is


def units_with_multiplier(value: Optional[UnitLike]) -> tuple["pint.Unit", float]:
    """Separate units and multiplier from :any:`.UnitLike`.

    Returns
    -------
    tuple
        1. :class:`pint.Unit`.
        2. :class:`float`; any multiplier on the units.
    """
    import pint

    registry = pint.get_application_registry()

    units = value or "1.0 dimensionless"
    if isinstance(units, str):
        uq = registry(units)
    elif isinstance(units, pint.Unit):
        uq = registry.Quantity(1.0, units)
    else:
        uq = units

    return uq.units, uq.magnitude


def unquote(value):
    """Reverse :func:`dask.core.quote`."""
    if isinstance(value, tuple) and len(value) == 1 and isinstance(value[0], literal):
        return value[0].data
    else:
        return value



"""Compatibility with :mod:`xarray`."""

from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Hashable,
    Iterable,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd
import xarray
from xarray.core import dtypes
from xarray.core.coordinates import Coordinates
from xarray.core.indexes import Indexes
from xarray.core.utils import either_dict_or_kwargs, is_scalar

if TYPE_CHECKING:
    import xarray.core.types
    from xarray.core.types import Dims

T = TypeVar("T", covariant=True)

__all__ = [
    "Coordinates",
    "DataArrayLike",
    "Indexes",
    "dtypes",
    "either_dict_or_kwargs",
    "is_scalar",
]


class DataArrayLike(Protocol):
    """Protocol for a :class:`.xarray.DataArray` -like API.

    This class is used to set signatures and types for methods and attributes on
    :class:`.AttrSeries` class, which then supplies implementations of each method.
    Objects typed :class:`.AnyQuantity` see either the signatures of this protocol, or
    identical signatures for the same methods on :class:`~xarray.DataArray` via
    :class:`.SparseDataArray`.
    """

    # Type hints for mypy in downstream applications
    @abstractmethod
    def __len__(self) -> int: ...

    def __mod__(self, other): ...
    def __mul__(self, other): ...
    def __neg__(self): ...
    def __pow__(self, other): ...
    def __radd__(self, other): ...
    def __rmul__(self, other): ...
    def __rsub__(self, other): ...
    def __rtruediv__(self, other): ...
    def __truediv__(self, other): ...

    @property
    @abstractmethod
    def data(self) -> Any: ...

    @property
    @abstractmethod
    def coords(self) -> xarray.core.coordinates.DataArrayCoordinates: ...

    @property
    @abstractmethod
    def dims(self) -> tuple[Hashable, ...]: ...

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]: ...

    @property
    @abstractmethod
    def size(self) -> int: ...

    def assign_coords(
        self,
        coords: Optional[Mapping[Any, Any]] = None,
        **coords_kwargs: Any,
    ): ...

    @abstractmethod
    def astype(
        self,
        dtype,
        *,
        order=None,
        casting=None,
        subok=None,
        copy=None,
        keep_attrs=True,
    ): ...

    @abstractmethod
    def bfill(
        self,
        dim: Hashable,
        limit: Optional[int] = None,
    ): ...

    @abstractmethod
    def clip(
        self,
        min: Optional["xarray.core.types.ScalarOrArray"] = None,
        max: Optional["xarray.core.types.ScalarOrArray"] = None,
        *,
        keep_attrs: Optional[bool] = None,
    ): ...

    @abstractmethod
    def copy(
        self,
        deep: bool = True,
        data: Any = None,
    ): ...

    @abstractmethod
    def cumprod(
        self,
        dim: "Dims" = None,
        *,
        skipna: Optional[bool] = None,
        keep_attrs: Optional[bool] = None,
        **kwargs: Any,
    ): ...

    @abstractmethod
    def drop_vars(
        self,
        names: Union[
            str, Iterable[Hashable], Callable[[Any], Union[str, Iterable[Hashable]]]
        ],
        *,
        errors="raise",
    ): ...

    @abstractmethod
    def expand_dims(
        self,
        dim=None,
        axis=None,
        **dim_kwargs: Any,
    ): ...

    @abstractmethod
    def ffill(
        self,
        dim: Hashable,
        limit: Optional[int] = None,
    ): ...

    @abstractmethod
    def groupby(
        self,
        group,
        squeeze: bool = True,
        restore_coord_dims: bool = False,
    ): ...

    @abstractmethod
    def interp(
        self,
        coords: Optional[Mapping[Any, Any]] = None,
        method: "xarray.core.types.InterpOptions" = "linear",
        assume_sorted: bool = False,
        kwargs: Optional[Mapping[str, Any]] = None,
        **coords_kwargs: Any,
    ): ...

    @abstractmethod
    def item(self, *args): ...

    @abstractmethod
    def max(
        self,
        dim: "Dims" = None,
        *,
        skipna: Optional[bool] = None,
        keep_attrs: Optional[bool] = None,
        **kwargs: Any,
    ): ...

    @abstractmethod
    def min(
        self,
        dim: "Dims" = None,
        *,
        skipna: Optional[bool] = None,
        keep_attrs: Optional[bool] = None,
        **kwargs: Any,
    ): ...

    @abstractmethod
    def pipe(
        self,
        func: Union[Callable[..., T], tuple[Callable[..., T], str]],
        *args: Any,
        **kwargs: Any,
    ): ...

    @abstractmethod
    def rename(
        self,
        new_name_or_name_dict: Union[Hashable, Mapping[Any, Hashable]] = None,
        **names: Hashable,
    ): ...

    @abstractmethod
    def round(self, *args, **kwargs): ...

    @abstractmethod
    def sel(
        self,
        indexers: Optional[Mapping[Any, Any]] = None,
        method: Optional[str] = None,
        tolerance=None,
        drop: bool = False,
        **indexers_kwargs: Any,
    ): ...

    @abstractmethod
    def shift(
        self,
        shifts: Optional[Mapping[Any, int]] = None,
        fill_value: Any = None,
        **shifts_kwargs: int,
    ): ...

    def squeeze(
        self,
        dim: Union[Hashable, Iterable[Hashable], None] = None,
        drop: bool = False,
        axis: Union[int, Iterable[int], None] = None,
    ): ...

    @abstractmethod
    def sum(
        self,
        dim: "Dims" = None,
        # Signature from xarray.DataArray
        *,
        skipna: Optional[bool] = None,
        min_count: Optional[int] = None,
        keep_attrs: Optional[bool] = None,
        **kwargs: Any,
    ): ...

    @abstractmethod
    def to_dataframe(
        self,
        name: Optional[Hashable] = None,
        dim_order: Optional[Sequence[Hashable]] = None,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def to_numpy(self) -> np.ndarray: ...

    @abstractmethod
    def where(self, cond: Any, other: Any = dtypes.NA, drop: bool = False): ...

    # Provided only for type-checking in other packages. AttrSeries implements;
    # SparseDataArray uses the xr.DataArray method.
    @abstractmethod
    def to_series(self) -> pd.Series: ...



import pytest

from genno import Computer, Operator


def null():
    pass  # pragma: no cover


class TestOperator:
    def test_add_tasks(self) -> None:
        op = Operator.define()(null)

        c = Computer()
        with pytest.raises(NotImplementedError):
            op.add_tasks(c)

    def test_add_tasks_deprecated(self) -> None:
        # Deprecated call triggers a warning
        with pytest.warns(DeprecationWarning):
            op = Operator.define(null)

        c = Computer()
        with pytest.raises(NotImplementedError):
            # Type of `op` is not Operator, so mypy complains
            op.add_tasks(c)  # type: ignore [attr-defined]



import re
from os import PathLike
from typing import Literal, Mapping, Optional, Union

from genno.core.describe import is_list_of_keys, label


def key_label(key):
    return unwrap(str(key))


_UNWRAP_EXPR = re.compile("^<(.*)>$")


def unwrap(label: str) -> str:
    """Unwrap any number of paired '<' and '>' at the start/end of `label`.

    These characters cause errors in graphviz/dot.
    """
    while True:
        result = _UNWRAP_EXPR.sub(r"\1", label)
        if result == label:
            return result
        else:
            label = result


class Visualizer:
    """Handle arguments for :func:`.visualize`."""

    def __init__(
        self,
        data_attributes: Mapping,
        function_attributes: Mapping,
        graph_attr: Mapping,
        node_attr: Mapping,
        edge_attr: Mapping,
        kwargs: Mapping,
    ):
        from graphviz import Digraph

        # Handle arguments
        self.da = data_attributes
        self.fa = function_attributes

        # Store for reference below
        self.ga = dict(graph_attr)
        self.ga.setdefault("rankdir", "BT")
        self.ga.update(kwargs)

        na = dict(node_attr)
        na.setdefault("fontname", "helvetica")

        # Create the graph and tracking collections
        self.graph = Digraph(graph_attr=self.ga, node_attr=na, edge_attr=edge_attr)

        # Nodes or edges already seen
        self.seen: set[str] = set()
        # Nodes already connected to the graph
        self.connected: set[str] = set()

    def get_attrs(self, kind: Literal["data", "func"], name: str, **defaults) -> dict:
        """Prepare attributes for a node of `kind`.

        If `name` is in self.da or self.fa, use those values, filling with `defaults`;
        otherwise, attributes are empty except for `defaults`."""
        if kind == "data":
            result = self.da.get(name, {}).copy()
            result.setdefault("shape", "ellipse")
        else:
            result = self.fa.get(name, {}).copy()
            # Use a directional shape like [> in LR mode; otherwise a box
            result.setdefault("shape", "cds" if self.ga["rankdir"] == "LR" else "box")

        [result.setdefault(k, v) for k, v in defaults.items()]

        return result

    def add_edge(self, a, b) -> None:
        """Add an edge to the graph."""
        self.graph.edge(a, b)
        # Update the connected nodes
        self.connected.update((a, b))

    def add_node(self, kind: Literal["data", "func"], name: str, k, v=None) -> None:
        """Add a data node to the graph."""
        if name in self.seen:
            return
        self.seen.add(name)
        _label = key_label(k) if kind == "data" else unwrap(label(v[0], max_length=50))
        self.graph.node(name, **self.get_attrs(kind, k, label=_label))

    def process(self, dsk: Mapping, collapse_outputs: bool):
        """Process the dask graph `dsk`."""
        from dask.core import get_dependencies, ishashable, istask
        from dask.dot import name

        # Iterate over keys, tasks in the graph
        for k, v in dsk.items():
            # A unique "name" for the node within `g`; similar to hash(k).
            k_name = name(k)

            if istask(v):  # A task
                # Node name for the operation, possibly distinct from its output
                func_name = name((k, "function")) if not collapse_outputs else k_name

                # Add a node for the operation
                self.add_node("func", func_name, k, v)

                # Add an edge between the operation-node and the key-node of its output
                if not collapse_outputs:
                    self.add_edge(func_name, k_name)

                # Add edges between the operation-node and the key-nodes for each of its
                # inputs
                for dep in get_dependencies(dsk, k):
                    dep_name = name(dep)
                    self.add_node("data", dep_name, dep)
                    self.add_edge(dep_name, func_name)
            elif ishashable(v) and v in dsk:  # Simple alias of k → v
                self.add_edge(name(v), k_name)
            elif is_list_of_keys(v, dsk):  # k = list of multiple keys (genno extension)
                for _v in v:
                    self.add_edge(name(_v), k_name)

            if not collapse_outputs or k_name in self.connected:
                # Something else that hasn't been seen: add a node that may never be
                # connected
                self.add_node("data", k_name, k)

        return self.graph


def visualize(
    dsk: Mapping,
    filename: Optional[Union[str, PathLike]] = None,
    format: Optional[str] = None,
    data_attributes: Optional[Mapping] = None,
    function_attributes: Optional[Mapping] = None,
    graph_attr: Optional[Mapping] = None,
    node_attr: Optional[Mapping] = None,
    edge_attr: Optional[Mapping] = None,
    collapse_outputs: bool = False,
    **kwargs,
):
    """Generate a Graphviz visualization of `dsk`.

    This is merged and extended version of :func:`dask.base.visualize`,
    :func:`dask.dot.dot_graph`, and :func:`dask.dot.to_graphviz` that produces
    informative output for genno graphs.

    Parameters
    ----------
    dsk :
        The graph to display.
    filename : Path or str, optional
        The name of the file to write to disk. If the file name does not have a suffix,
        ".png" is used by default. If `filename` is :data:`None`, no file is written,
        and dask communicates with :program:`dot` using only pipes.
    format : {'png', 'pdf', 'dot', 'svg', 'jpeg', 'jpg'}, optional
        Format in which to write output file, if not given by the suffix of `filename`.
        Default "png".
    data_attributes :
        Graphviz attributes to apply to single nodes representing keys, in addition to
        `node_attr`.
    function_attributes :
        Graphviz attributes to apply to single nodes representing operations or
        functions, in addition to `node_attr`.
    graph_attr :
        Mapping of (attribute, value) pairs for the graph. Passed directly to
        :class:`.graphviz.Digraph`.
    node_attr :
        Mapping of (attribute, value) pairs set for all nodes. Passed directly to
        :class:`.graphviz.Digraph`.
    edge_attr :
        Mapping of (attribute, value) pairs set for all edges. Passed directly to
        :class:`.graphviz.Digraph`.
    collapse_outputs : bool, optional
        Omit nodes for keys that are the output of intermediate calculations.
    kwargs :
        All other keyword arguments are added to `graph_attr`.

    Examples
    --------

    .. _visualize-example:

    Prepare a computer:

    >>> from genno import Computer
    >>> from genno.testing import add_test_data
    >>> c = Computer()
    >>> add_test_data(c)
    >>> c.add_product("z", "x:t", "x:y")
    >>> c.add("y::0", itemgetter(0), "y")
    >>> c.add("y0", "y::0")
    >>> c.add("index_to", "z::indexed", "z:y", "y::0")
    >>> c.add_single("all", ["z::indexed", "t", "config", "x:t"])

    Visualize its contents:

    >>> c.visualize("example.svg")

    This produces the output:

    .. image:: _static/visualize.svg
       :alt: Example output from graphviz.visualize.

    See also
    --------
    .describe.label
    """
    from dask.dot import graphviz_to_file

    # Handle arguments
    v = Visualizer(
        data_attributes or {},
        function_attributes or {},
        graph_attr or {},
        node_attr or {},
        edge_attr or {},
        kwargs,
    )

    # Process the graph
    graph = v.process(dsk, collapse_outputs)

    return graphviz_to_file(graph, None if filename is None else str(filename), format)



from typing import Optional, Union

import sdmx


def handle_od(
    value: Union[str, "sdmx.model.common.DimensionComponent", None],
    structure: "sdmx.model.common.BaseDataStructureDefinition",
) -> Optional["sdmx.model.common.DimensionComponent"]:
    """Handle `observation_dimension` arguments for :mod:`.sdmx.operator`.

    Ensure either None or a DimensionComponent.
    """
    import sdmx

    if isinstance(value, sdmx.model.common.DimensionComponent) or value is None:
        return value
    elif value is not None:
        return structure.dimensions.get(value)


def urn(obj: "sdmx.model.common.MaintainableArtefact") -> str:
    """Return the URN of `obj`, or construct it."""
    if result := obj.urn:  # pragma: no cover
        return result
    else:
        return sdmx.urn.make(obj)


def handle_version(
    version: Union["sdmx.format.Version", str, None],
) -> tuple[
    "sdmx.format.Version",
    type["sdmx.model.common.BaseDataSet"],
    type["sdmx.model.common.BaseObservation"],
]:
    """Handle `version` arguments for :mod:`.sdmx.operator`.

    Also return either :mod:`sdmx.model.v21` or :mod:`sdmx.model.v30`, as appropriate.
    """
    from sdmx.format import Version

    # Ensure a Version enum member
    if not isinstance(version, Version):
        version = Version[version or "2.1"]

    # Retrieve information model module
    im = {Version["2.1"]: sdmx.model.v21, Version["3.0.0"]: sdmx.model.v30}[version]

    return (
        version,
        im.get_class("StructureSpecificDataSet"),
        im.get_class("Observation"),
    )



import logging
from functools import partial
from itertools import tee
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Hashable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Union,
    cast,
)

import numpy as np
import pandas as pd
import pandas.core.indexes.base as ibase
import xarray as xr
from packaging.version import Version
from pandas.core.generic import NDFrame
from pandas.core.internals.base import DataManager

from genno.compat.pandas import version as pandas_version
from genno.compat.xarray import (
    Coordinates,
    DataArrayLike,
    Indexes,
    dtypes,
    either_dict_or_kwargs,
    is_scalar,
)

from .base import BaseQuantity, collect_attrs, rank, single_column_df

if TYPE_CHECKING:
    from _typeshed import SupportsRichComparisonT

    from genno.types import Dims


log = logging.getLogger(__name__)


def _ensure_multiindex(obj):
    """Ensure `obj` has a pd.MultiIndex, even if 1D."""
    try:
        obj.index.levels  # Access an Attribute of MultiIndex that Index does not have
    except AttributeError:
        # Assign the dimension name "dim_0" if 1-D with no names
        kw = {}
        if len(obj.index) > 1 and obj.index.name is None:
            kw["names"] = ["dim_0"]
        obj.index = pd.MultiIndex.from_product([obj.index], **kw)
    return obj


class AttrSeriesCoordinates(Coordinates):
    def __init__(self, obj):
        self._data = obj
        self._idx = obj.index.remove_unused_levels()

    @property
    def _names(self):
        return tuple(filter(None, self._idx.names))

    @property
    def variables(self):
        result = {}
        for name, levels in zip(self._idx.names, self._idx.levels):
            if name is None:
                continue
            result[name] = levels.unique()
        return result

    def __contains__(self, key: Hashable) -> bool:
        return key in self._names

    def __getitem__(self, key):
        levels = self._idx.levels[self._idx.names.index(key)].to_list()
        return xr.DataArray(levels, coords={key: levels})


class AttrSeries(BaseQuantity, pd.Series, DataArrayLike):
    """:class:`pandas.Series` subclass imitating :class:`xarray.DataArray`.

    The AttrSeries class provides similar methods and behaviour to
    :class:`xarray.DataArray`, so that :mod:`genno.operator` functions and user code can
    use xarray-like syntax. In particular, this allows such code to be agnostic about
    the order of dimensions.

    Parameters
    ----------
    units : str or pint.Unit, optional
        Set the units attribute. The value is converted to :class:`pint.Unit` and added
        to `attrs`.
    attrs : :class:`~collections.abc.Mapping`, optional
        Set the :attr:`~pandas.Series.attrs` of the AttrSeries. This attribute was added
        in `pandas 1.0 <https://pandas.pydata.org/docs/whatsnew/v1.0.0.html>`_, but is
        not currently supported by the Series constructor.
    """

    # See https://pandas.pydata.org/docs/development/extending.html
    @property
    def _constructor(self):
        return AttrSeries

    def __init__(
        self,
        data: Any = None,
        *args,
        name: Optional[Hashable] = None,
        attrs: Optional[Mapping] = None,
        **kwargs,
    ):
        # Emulate behaviour of Series.__init__
        if isinstance(data, DataManager) and "fastpath" not in kwargs:
            if not (
                0 == len(args) == len(kwargs) and attrs is None
            ):  # pragma: no cover
                raise NotImplementedError
            NDFrame.__init__(self, data)
            if name:
                self.name = name
            return

        attrs = collect_attrs(data, attrs, kwargs)

        if isinstance(data, (pd.Series, xr.DataArray)):
            # Extract name from existing object or use the argument
            name = ibase.maybe_extract_name(name, data, type(self))

            try:
                # Pre-convert from xr.DataArray to pd.Series to preserve names and
                # labels. For AttrSeries, this is a no-op (see below).
                data = data.to_series()
            except AttributeError:
                # pd.Series
                pass
            except ValueError:
                # xr.DataArray
                if data.shape == tuple():
                    # data is a scalar/0-dimensional xr.DataArray. Pass the 1 value
                    data = data.data
                else:  # pragma: no cover
                    raise

        data, name = single_column_df(data, name)

        if data is None:
            kwargs["dtype"] = float
        elif coords := kwargs.pop("coords", None):
            # Handle xarray-style coords arg
            data = np.array(data).ravel()
            kwargs["index"] = pd.MultiIndex.from_product(
                list(coords.values()), names=list(coords.keys())
            )

        # Don't pass attrs to pd.Series constructor; it currently does not accept them
        pd.Series.__init__(self, data, *args, name=name, **kwargs)

        # Ensure a MultiIndex
        _ensure_multiindex(self)

        # Update the attrs after initialization
        self._attrs.update(attrs)

    def __repr__(self):
        return (
            super().__repr__() + f", units: {self.attrs.get('_unit', 'dimensionless')}"
        )

    @classmethod
    def from_series(cls, series, sparse=None):
        """Like :meth:`xarray.DataArray.from_series`."""
        return AttrSeries(series)

    @staticmethod
    def _perform_binary_op(
        op, left: "AttrSeries", right: "AttrSeries", factor: float
    ) -> "AttrSeries":
        # Ensure both operands are multi-indexed, and have at least 1 common dim
        if left.dims:
            order, right = right.align_levels(left)
        else:
            order, left = left.align_levels(right)

        # Invoke a pd.Series method like .mul()
        fv = dict(fill_value=0.0) if rank(op) == 1 else {}
        return getattr(left, op.__name__)(right, **fv).dropna().reorder_levels(order)

    def assign_coords(self, coords=None, **coord_kwargs):
        """Like :meth:`xarray.DataArray.assign_coords`."""
        coords = either_dict_or_kwargs(coords, coord_kwargs, "assign_coords")

        # Construct a new index
        new_idx = self.index.copy()
        for dim, values in coords.items():
            expected_len = len(self.index.levels[self.index.names.index(dim)])
            if expected_len != len(values):
                raise ValueError(
                    f"conflicting sizes for dimension {repr(dim)}: length "
                    f"{expected_len} on <this-array> and length {len(values)} on "
                    f"{repr(dim)}"
                )

            new_idx = new_idx.set_levels(values, level=dim)

        # Return a new object with the new index
        return self.set_axis(new_idx)

    def bfill(self, dim: Hashable, limit: Optional[int] = None):
        """Like :meth:`xarray.DataArray.bfill`."""
        # TODO this likely does not work for 1-D quantities due to unstack(); test and
        #      if needed use _maybe_groupby()
        return self._replace(
            self.unstack(dim)
            .bfill(axis=1, limit=limit)
            .stack()
            .reorder_levels(self.dims),
        )

    @property
    def coords(self):
        """Like :attr:`xarray.DataArray.coords`. Read-only."""
        return AttrSeriesCoordinates(self)

    def cumprod(self, dim=None, axis=None, skipna=None, **kwargs):
        """Like :meth:`xarray.DataArray.cumprod`."""
        if axis:
            log.info(f"{self.__class__.__name__}.cumprod(…, axis=…) is ignored")
        if skipna is None:
            skipna = self.dtype == float
        if dim in (None, "..."):
            if len(self.dims) > 1:
                raise NotImplementedError("cumprod() over multiple dimensions")
            dim = self.dims[0]

        def _(s):
            # Invoke cumprod from the parent class pd.Series
            return super(pd.Series, s).cumprod(skipna=skipna, **kwargs)

        return self._replace(self._groupby_apply(dim, sorted(self.coords[dim].data), _))

    @property
    def data(self):
        return self.values

    @property
    def dims(self) -> tuple[Hashable, ...]:
        """Like :attr:`xarray.DataArray.dims`."""
        # If 0-D, the single dimension has name `None` → discard
        return tuple(filter(None, self.index.names))

    @property
    def shape(self) -> tuple[int, ...]:
        """Like :attr:`xarray.DataArray.shape`."""
        idx = self.index.remove_unused_levels()
        return tuple(len(idx.levels[i]) for i in map(idx.names.index, self.dims))

    def clip(
        self,
        min=None,
        max=None,
        *,
        keep_attrs: Optional[bool] = None,
    ):
        """Like :meth:`.xarray.DataArray.clip`.

        :meth:`.pandas.Series.clip` has arguments named `lower` and `upper` instead of
        `min` and `max`, respectively.

        :py:`keep_attrs=False` is not implemented.
        """
        if keep_attrs is False:
            raise NotImplementedError("clip(…, keep_attrs=False)")

        if pandas_version() < Version("2.1.0"):
            return self._replace(pd.Series(self).clip(min, max))
        else:
            return super().clip(min, max)  # type: ignore [safe-super]

    def drop(self, label):
        """Like :meth:`xarray.DataArray.drop`."""
        return self.droplevel(label)

    def drop_vars(
        self, names: Union[Hashable, Iterable[Hashable]], *, errors: str = "raise"
    ):
        """Like :meth:`xarray.DataArray.drop_vars`."""

        return self.droplevel(names)

    def expand_dims(self, dim=None, axis=None, **dim_kwargs: Any) -> "AttrSeries":
        """Like :meth:`xarray.DataArray.expand_dims`.

        .. todo:: Support passing a mapping of length > 1 to `dim`.
        """
        if isinstance(dim, list):
            dim = dict.fromkeys(dim, [])
        dim = either_dict_or_kwargs(dim, dim_kwargs, "expand_dims")
        if axis is not None:
            raise NotImplementedError  # pragma: no cover

        result = self
        for name, values in reversed(list(dim.items())):
            N = len(values)
            if N == 0:  # Dimension without labels
                N, values = 1, [None]
            result = pd.concat([result] * N, keys=values, names=[name], sort=False)

        # Ensure `result` is multiindexed
        try:
            i = result.index.names.index(None)
        except ValueError:
            pass
        else:
            assert 2 == len(result.index.names)
            result.index = pd.MultiIndex.from_product([result.index.droplevel(i)])

        return result

    def ffill(self, dim: Hashable, limit: Optional[int] = None):
        """Like :meth:`xarray.DataArray.ffill`."""
        # TODO this likely does not work for 1-D quantities due to unstack(); test and
        #      if needed use _maybe_groupby()
        return self._replace(
            self.unstack(dim)
            .ffill(axis=1, limit=limit)
            .stack()
            .reorder_levels(self.dims),
        )

    def item(self, *args):
        """Like :meth:`xarray.DataArray.item`."""
        if len(args) and args != (None,):
            raise NotImplementedError
        elif self.size != 1:
            raise ValueError
        return self.iloc[0]

    def interp(
        self,
        coords: Optional[Mapping[Hashable, Any]] = None,
        method: str = "linear",
        assume_sorted: bool = True,
        kwargs: Optional[Mapping[str, Any]] = None,
        **coords_kwargs: Any,
    ):
        """Like :meth:`xarray.DataArray.interp`.

        This method works around two long-standing bugs in :mod:`pandas`:

        - `pandas-dev/pandas#25460 <https://github.com/pandas-dev/pandas/issues/25460>`_
        - `pandas-dev/pandas#31949 <https://github.com/pandas-dev/pandas/issues/31949>`_
        """
        from scipy.interpolate import interp1d

        if kwargs is None:
            kwargs = {}

        coords = either_dict_or_kwargs(coords, coords_kwargs, "interp")
        if len(coords) > 1:
            raise NotImplementedError("interp() on more than 1 dimension")

        # Unpack the dimension and levels (possibly overlapping with existing)
        dim = list(coords.keys())[0]
        levels = coords[dim]
        # Ensure a list
        if isinstance(levels, (int, float)):
            levels = [levels]

        def _flat_index(obj: AttrSeries):
            """Unpack a 1-D MultiIndex from an AttrSeries."""
            return [v[0] for v in obj.index]

        # Group by `dim` so that each level appears ≤ 1 time in `group_series`

        def _(s):
            # Work around https://github.com/pandas-dev/pandas/issues/31949
            # Location of existing values
            x = s.notna()

            # Create an interpolator from the existing values
            i = interp1d(_flat_index(s[x]), s[x], kind=method, **kwargs)

            return s.fillna(pd.Series(i(_flat_index(s[~x])), index=s[~x].index))

        result = self._groupby_apply(dim, levels, _)

        # - Restore dimension order and attributes.
        # - Select only the desired `coords`.
        return self._replace(result.reorder_levels(self.dims)).sel(coords)

    def rename(
        self,
        new_name_or_name_dict: Union[Hashable, Mapping[Hashable, Hashable]] = None,
        **names: Hashable,
    ):
        """Like :meth:`xarray.DataArray.rename`."""
        if new_name_or_name_dict is None or isinstance(new_name_or_name_dict, Mapping):
            index = either_dict_or_kwargs(new_name_or_name_dict, names, "rename")
            return self.rename_axis(index=index)
        else:
            return self._set_name(new_name_or_name_dict)

    def sel(
        self,
        indexers: Optional[Mapping[Any, Any]] = None,
        method: Optional[str] = None,
        tolerance=None,
        drop: bool = False,
        **indexers_kwargs: Any,
    ):
        """Like :meth:`xarray.DataArray.sel`."""
        if method is not None:
            raise NotImplementedError(f"AttrSeries.sel(…, method={method!r})")
        if tolerance is not None:
            raise NotImplementedError(f"AttrSeries.sel(…, tolerance={tolerance!r})")

        indexers = either_dict_or_kwargs(indexers, indexers_kwargs, "sel")

        if len(indexers) and all(
            isinstance(i, xr.DataArray) for i in indexers.values()
        ):
            # DataArray indexers

            # Combine indexers in a data set; dimensions are aligned
            ds = xr.Dataset(indexers)

            # All dimensions indexed
            dims_indexed = set(indexers.keys())
            # Dimensions to discard
            dims_drop = set(ds.data_vars.keys())

            # Check contents of indexers
            if any(ds.isnull().any().values()):
                raise IndexError(
                    f"Dimensions of indexers mismatch: {ds.notnull().sum()}"
                )
            elif len(ds.dims) > 1:
                raise NotImplementedError(  # pragma: no cover
                    f"map to > 1 dimensions {repr(ds.dims)} with AttrSeries.sel()"
                )

            # pd.Index object with names and levels of the new dimension to be created
            idx = ds.coords.to_index()

            # Dimensions to drop on sliced data to avoid duplicated dimensions
            drop_slice = list(dims_indexed - dims_drop)

            # Dictionary of Series to concatenate
            series = {}

            # Iterate over labels in the new dimension
            for label in idx:
                # Get a slice from the indexers corresponding to this label
                loc_ds = ds.sel({idx.name: label})

                # Assemble a key with one element for each dimension
                seq0 = [loc_ds.get(d) for d in self.dims]
                # Replace None from .get() with slice(None) or unpack a single value
                seq1 = [slice(None) if item is None else item.item() for item in seq0]

                # Use the key to retrieve 1+ integer locations; slice; store
                series[label] = self.iloc[self.index.get_locs(seq1)].droplevel(
                    drop_slice
                )

            # Rejoin to a single data frame; drop the source levels
            data = pd.concat(series, names=[idx.name]).droplevel(list(dims_drop))
        else:
            # Other indexers

            # Iterate over dimensions
            idx = []
            to_drop = set()
            for dim in self.dims:
                # Get an indexer for this dimension
                i = indexers.get(dim, slice(None))

                if is_scalar(i) and (i != slice(None)):
                    to_drop.add(dim)

                # Maybe unpack an xarray DataArray indexers, for pandas
                idx.append(i.data if isinstance(i, xr.DataArray) else i)

            # Select
            data = self.loc[tuple(idx)]

            # Only drop if not returning a scalar value
            if isinstance(data, pd.Series):
                # Drop levels where a single value was selected
                data = data.droplevel(list(to_drop & set(data.index.names)))

        # Return
        return self._replace(data)

    def shift(
        self,
        shifts: Optional[Mapping[Hashable, int]] = None,
        fill_value: Any = None,
        **shifts_kwargs: int,
    ):
        """Like :meth:`xarray.DataArray.shift`."""
        shifts = either_dict_or_kwargs(shifts, shifts_kwargs, "shift")

        # Apply shifts one-by-one
        result = self
        for dim, periods in shifts.items():
            levels = sorted(self.coords[dim].data)

            def _(s):
                # Invoke shift from the parent class pd.Series
                return super(pd.Series, s).shift(periods=periods, fill_value=fill_value)

            result = result._groupby_apply(dim, levels, _)

        return self._replace(result)

    def sum(
        self,
        dim: "Dims" = None,
        # Signature from xarray.DataArray
        # *,
        skipna: Optional[bool] = None,
        min_count: Optional[int] = None,
        keep_attrs: Optional[bool] = None,
        **kwargs: Any,
    ) -> "AttrSeries":
        """Like :meth:`xarray.DataArray.sum`."""
        if skipna is not None or min_count is not None:
            raise NotImplementedError

        if dim is None or isinstance(dim, Hashable):
            dim = tuple(filter(None, (dim,)))

        # Check dimensions
        bad_dims = set(dim) - set(self.index.names)
        if bad_dims:
            raise ValueError(
                f"{bad_dims} not found in array dimensions {self.index.names}"
            )

        # Create the object on which to .sum()
        return self._replace(self._maybe_groupby(dim).sum(**kwargs))

    def squeeze(self, dim=None, drop=False, axis=None):
        """Like :meth:`xarray.DataArray.squeeze`."""

        idx = self.index.remove_unused_levels()

        to_drop = []
        for i, name in enumerate(filter(None, idx.names)):
            if dim and name != dim:
                continue
            elif len(idx.levels[i]) > 1:
                if dim is None:
                    continue
                else:
                    raise ValueError(
                        "cannot select a dimension to squeeze out which has length "
                        "greater than one"
                    )

            to_drop.append(name)

        if dim and not to_drop:
            raise KeyError(dim)  # Specified dimension does not exist

        if set(to_drop) == set(self.dims):
            # Dropping all dimensions → 0-D quantity; simply reset
            return self.reset_index(drop=True)
        else:
            return self.droplevel(to_drop).pipe(_ensure_multiindex)

    def transpose(self, *dims):
        """Like :meth:`xarray.DataArray.transpose`."""
        return self.reorder_levels(dims)

    def to_dataframe(
        self,
        name: Optional[Hashable] = None,
        dim_order: Optional[Sequence[Hashable]] = None,
    ) -> pd.DataFrame:
        """Like :meth:`xarray.DataArray.to_dataframe`."""
        if dim_order is not None:
            raise NotImplementedError("dim_order arg to to_dataframe()")

        self.name = name or self.name or "value"  # type: ignore
        return self.to_frame()

    def to_series(self):
        """Like :meth:`xarray.DataArray.to_series`."""
        return self

    def where(
        self,
        cond: Any,
        other: Any = dtypes.NA,
        drop: bool = False,
        *,
        axis=None,  # Needed internally to pd.Series.clip()
        inplace: bool = False,  # Needed internally to pd.Series.clip()
    ):
        """Like :meth:`xarray.DataArray.where`.

        Passing :any:`True` for `drop` is not implemented.
        """
        if drop is True:
            raise NotImplementedError("where(…, drop=True)")
        elif axis is not None or inplace is not False:
            raise NotImplementedError("where(…, axis=…) or where(…, inplace=…)")
        return super().where(cond, other)  # type: ignore [safe-super]

    @property
    def xindexes(self):  # pragma: no cover
        # NB incomplete implementation; currently sufficient that this property exists
        return Indexes(dict(), None)

    # Internal methods
    def align_levels(
        self, other: "AttrSeries"
    ) -> tuple[Sequence[Hashable], "AttrSeries"]:
        """Return a copy of `self` with ≥1 dimension(s) in the same order as `other`.

        Work-around for https://github.com/pandas-dev/pandas/issues/25760 and other
        limitations of :class:`pandas.Series`.
        """
        # Union of dimensions of `self` and `other`; initially just other
        d_union = list(other.dims)

        # Lists of common dimensions, and dimensions on `other` missing from `self`.
        d_common = []  # Common dimensions of `self` and `other`
        d_other_only = []  # (dimension, index) of `other` missing from `self`
        for i, d in enumerate(d_union):
            if d in self.index.names:
                d_common.append(d)
            else:
                d_other_only.append((d, i))

        result = self
        d_result = []  # Order of dimensions on the result

        if len(d_common) == 0:
            # No common dimensions between `other` and `self`
            if len(d_other_only):
                # …but `other` is ≥1D
                # Broadcast the result over the final missing dimension of `other`
                d, i = d_other_only[-1]
                result = result.expand_dims({d: other.index.levels[i]})
                # Reordering starts with this dimension
                d_result.append(d)
            elif not result.dims:
                # Both `self` and `other` are scalar
                d_result.append(None)
        else:
            # Some common dimensions exist; no need to broadcast, only reorder
            d_result.extend(d_common)

        # Append the dimensions of `self`
        i1, i2 = tee(filter(lambda n: n not in other.dims, self.dims), 2)
        d_union.extend(i1)
        d_result.extend(i2)

        return d_union or [None], result.reorder_levels(d_result or [None])

    def _groupby_apply(
        self,
        dim: Hashable,
        levels: Iterable["SupportsRichComparisonT"],
        func: Callable[["AttrSeries"], "AttrSeries"],
    ) -> "AttrSeries":
        """Group along `dim`, ensure levels `levels`, and apply `func`.

        `func` should accept and return AttrSeries. The resulting AttrSeries are
        concatenated again along `dim`.
        """
        # Preserve order of dimensions
        dims = self.dims

        # Dimension other than `dim`
        d_other = list(filter(lambda d: d != dim, dims))

        def _join(base, item):
            """Rejoin a full key for the MultiIndex in the correct order."""
            # Wrap a scalar `base` (only occurs with len(other_dims) == 1; pandas < 2.0)
            base = list(base) if isinstance(base, tuple) else [base]
            return [(base[d_other.index(d)] if d in d_other else item[0]) for d in dims]

        # Grouper or iterable of (key, pd.Series)
        groups = self.groupby(d_other) if len(d_other) else [(None, self)]

        # Iterate over groups, accumulating modified series
        result = []
        for group_key, group_series in groups:
            # Work around https://github.com/pandas-dev/pandas/issues/25460; can't do:
            # group_series.reindex(…, level=dim)

            # Create 1-D MultiIndex for `dim` with the union of existing coords and
            # `levels`
            _levels = set(levels)
            _levels.update(group_series.index.get_level_values(dim))
            idx = pd.MultiIndex.from_product([sorted(_levels)], names=[dim])
            # Reassemble full MultiIndex with the new coords added along `dim`
            full_idx = pd.MultiIndex.from_tuples(
                map(partial(_join, group_key), idx), names=dims
            )

            # - Reindex with `full_idx` to insert NaNs for new `levels`.
            # - Replace the with the 1D index for `dim` only.
            # - Apply `func`.
            # - Restore the full index.
            result.append(
                func(group_series.reindex(full_idx).set_axis(idx)).set_axis(full_idx)
            )

        return pd.concat(result)

    def _maybe_groupby(self, dim):
        """Return an object for operations along dimension(s) `dim`.

        If `dim` is a subset of :attr:`dims`, returns a SeriesGroupBy object along the
        other dimensions.
        """
        if len(set(dim)) in (0, len(self.index.names)):
            return cast(pd.Series, super())
        else:
            # Group on dimensions other than `dim`
            levels = list(filter(lambda d: d not in dim, self.index.names))
            return self.groupby(level=levels, group_keys=False, observed=True)

    def _replace(self, data) -> "AttrSeries":
        """Shorthand to preserve attrs."""
        return self.__class__(data, name=self.name, attrs=self.attrs)



import logging
from contextlib import contextmanager
from functools import lru_cache

from packaging.version import Version, parse

log = logging.getLogger(__name__)


@lru_cache
def version() -> Version:
    import pandas

    return parse(pandas.__version__)


@contextmanager
def disable_copy_on_write(name):
    """Context manager to disable Pandas :ref:`pandas:copy_on_write`.

    A message is logged with level :any:`logging.DEBUG` if the setting is changed.
    """
    import pandas

    stored = pandas.options.mode.copy_on_write
    override_value = "warn" if version() >= Version("2.2.0") else False

    try:
        if stored is True:
            log.debug(f"Override pandas.mode.options.copy_on_write = True for {name}")
            pandas.options.mode.copy_on_write = override_value
        yield
    finally:
        pandas.options.mode.copy_on_write = stored


@lru_cache
def handles_parquet_attrs() -> bool:
    """Return :any:`True` if :mod:`pandas` can read/write attrs to/from Parquet files.

    If not, a message is logged.
    """
    if version() < Version("2.1.0"):
        log.info(
            f"Pandas {version()!s} < 2.1.0 cannot read/write Quantity.attrs "
            f"to/from Parquet; {__name__} will use pickle from the standard library"
        )
        return False
    else:
        return True



from typing import Hashable, Iterable, Mapping, Optional, Union

import genno
from genno import Quantity

try:
    import sdmx
except ModuleNotFoundError:  # pragma: no cover
    HAS_SDMX = False
else:
    HAS_SDMX = True

from . import util

__all__ = [
    "codelist_to_groups",
    "dataset_to_quantity",
    "quantity_to_dataset",
    "quantity_to_message",
]


def codelist_to_groups(
    codes: Union["sdmx.model.common.Codelist", Iterable["sdmx.model.common.Code"]],
    dim: Optional[str] = None,
) -> Mapping[str, Mapping[str, list[str]]]:
    """Convert `codes` into a mapping from parent items to their children.

    The returned value is suitable for use with :func:`~.operator.aggregate`.

    Parameters
    ----------
    codes
        Either a :class:`sdmx.Codelist <sdmx.model.common.Codelist>` object or any
        iterable of :class:`sdmx.Code <sdmx.model.common.Code>`.
    dim : str, optional
        Dimension to aggregate. If `codes` is a code list and `dim` is not given, the
        ID of the code list is used; otherwise `dim` must be supplied.
    """
    from sdmx.model.common import Codelist

    if isinstance(codes, Codelist):
        items: Iterable["sdmx.model.common.Code"] = codes.items.values()
        dim = dim or codes.id
    else:
        items = codes

    if dim is None:
        raise ValueError("Must provide a dimension ID for aggregation")

    groups = dict()
    for code in filter(lambda c: len(c.child), items):
        groups[code.id] = list(map(str, code.child))

    return {dim: groups}


def dataset_to_quantity(ds: "sdmx.model.common.BaseDataSet") -> Quantity:
    """Convert :class:`DataSet <sdmx.model.common.BaseDataSet>` to :class:`.Quantity`.

    Returns
    -------
    .Quantity
        The quantity may have the attributes:

        - "dataflow_urn": :attr:`urn <sdmx.model.common.IdentifiableArtefact.urn>` of
          the :class:`Dataflow <sdmx.model.common.BaseDataflow>` referenced by the
          :attr:`described_by <sdmx.model.common.BaseDataSet.described_by>` attribute of
          `ds`, if any.
        - "structure_urn": :attr:`urn <sdmx.model.common.IdentifiableArtefact.urn>` of
          the :class:`DataStructureDefinition
          <sdmx.model.common.BaseDataStructureDefinition>` referenced by the
          :attr:`structured_by <sdmx.model.common.BaseDataSet.structured_by>` attribute
          of `ds`, if any.
    """
    # Assemble attributes
    attrs: dict[str, str] = {}
    if ds.described_by:  # pragma: no cover
        attrs.update(dataflow_urn=util.urn(ds.described_by))
    if ds.structured_by:
        attrs.update(structure_urn=util.urn(ds.structured_by))

    return Quantity(sdmx.to_pandas(ds), attrs=attrs)


def quantity_to_dataset(
    qty: Quantity,
    structure: "sdmx.model.common.BaseDataStructureDefinition",
    *,
    observation_dimension: Optional[str] = None,
    version: Union["sdmx.format.Version", str, None] = None,
) -> "sdmx.model.common.BaseDataSet":
    """Convert :class:`.Quantity` to :class:`DataSet <sdmx.model.common.BaseDataSet>`.

    The resulting data set is structure-specific.

    Parameters
    ----------
    observation_dimension : str or sdmx.model.common.DimensionComponent, optional
        If given, the resulting data set is arranged in series, with the
        `observation_dimension` varying across observations within each series. If not
        given, the data set is flat, with all dimensions specified for each observation.
    version : str or sdmx.format.Version, optional
        SDMX data model version to use; default 2.1.
    """
    # Handle `version` argument, identify classes
    _, DataSet, Observation = util.handle_version(version)
    Key = sdmx.model.common.Key
    SeriesKey = sdmx.model.common.SeriesKey

    # Narrow type
    # NB This is necessary because BaseDataStructureDefinition.measures is not defined
    # TODO Remove once addressed upstream
    assert isinstance(
        structure,
        (
            sdmx.model.v21.DataStructureDefinition,
            sdmx.model.v30.DataStructureDefinition,
        ),
    )

    try:
        # URN of DSD stored on `qty` matches `structure`
        assert qty.attrs["structure_urn"] == util.urn(structure)
    except KeyError:
        pass  # No such attribute

    # Dimensions; should be equivalent to the IDs of structure.dimensions
    dims = qty.dims

    # Create data set
    ds = DataSet(structured_by=structure)
    measure = structure.measures[0]

    if od := util.handle_od(observation_dimension, structure):
        # Index of `observation_dimension`
        od_index = dims.index(od.id)
        # Group data / construct SeriesKey all *except* the observation_dimension
        series_dims = list(dims[:od_index] + dims[od_index + 1 :])
        grouped: Iterable = qty.to_series().groupby(series_dims)
        # For as_obs()
        obs_dims: tuple[Hashable, ...] = (od.id,)
        key_slice = slice(od_index, od_index + 1)
    else:
        # Pseudo-groupby object
        grouped = [(None, qty.to_series())]
        obs_dims, key_slice = dims, slice(None)

    def as_obs(key, value):
        """Convert a single pd.Series element to an sdmx Observation."""
        return Observation(
            # Select some or all elements of the SeriesGroupBy key
            dimension=structure.make_key(Key, dict(zip(obs_dims, key[key_slice]))),
            value_for=measure,
            value=value,
        )

    for series_key, data in grouped:
        if series_key:
            sk = structure.make_key(SeriesKey, dict(zip(series_dims, series_key)))
        else:
            sk = None

        # - Convert each item to an sdmx Observation.
        # - Add to `ds`, associating with sk
        ds.add_obs([as_obs(key, value) for key, value in data.items()], series_key=sk)

    return ds


def quantity_to_message(
    qty: Quantity, structure: "sdmx.model.v21.DataStructureDefinition", **kwargs
) -> "sdmx.message.DataMessage":
    """Convert :class:`.Quantity` to :class:`DataMessage <sdmx.message.DataMessage>`.

    Parameters
    ----------
    kwargs :
        `observation_dimension` and `version` parameters are both used and passed on
        to :func:`.quantity_to_dataset`.
    """
    kwargs.update(
        version=util.handle_version(kwargs.get("version"))[0],
        observation_dimension=util.handle_od(
            kwargs.get("observation_dimension"), structure
        ),
    )

    ds = quantity_to_dataset(
        qty,
        structure,
        observation_dimension=kwargs["observation_dimension"],
        version=kwargs["version"],
    )

    return sdmx.message.DataMessage(data=[ds], **kwargs)


@genno.operator.write_report.register
def _(obj: "sdmx.message.DataMessage", path, kwargs=None) -> None:
    """Write  `obj` to the file at `path`.

    If `obj` is a :class:`sdmx.message.DataMessage` and `path` ends with ".xml", use
    use :mod:`sdmx` methods to write the file to SDMX-ML. Otherwise, equivalent to
    :func:`genno.operator.write_report`.
    """
    import genno.compat.sdmx.operator  # noqa: F401

    assert path.suffix.lower() == ".xml"

    kwargs = kwargs or {}
    kwargs.setdefault("pretty_print", True)

    path.write_bytes(sdmx.to_xml(obj, **kwargs))



import logging
import re
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Collection,
    Iterable,
    Mapping,
    Optional,
    Union,
)
from warnings import warn

import pyam

import genno
import genno.operator
from genno.core.key import Key, KeyLike
from genno.core.operator import Operator

from . import util

if TYPE_CHECKING:
    import pandas

    from genno.core.computer import Computer
    from genno.core.quantity import AnyQuantity

log = logging.getLogger(__name__)


__all__ = [
    "as_pyam",
    "quantity_from_iamc",
]


@Operator.define()
def as_pyam(
    scenario,
    quantity: "AnyQuantity",
    *,
    rename: Optional[Mapping[str, str]] = None,
    collapse: Optional[Callable] = None,
    replace=dict(),
    drop: Union[Collection[str], str] = "auto",
    unit=None,
    prepend_name: bool = True,
    model_name: Optional[str] = None,
    scenario_name: Optional[str] = None,
):
    """Return a :class:`pyam.IamDataFrame` containing the data from `quantity`.

    Warnings are logged if the arguments result in additional, unhandled columns in the
    resulting data frame that are not part of the IAMC spec.

    The conversion has the following steps:

    1. `quantity` is converted to a temporary :class:`pandas.DataFrame`.
    2. Labels for the following IAMC dimensions are filled:

       - ``model``, ``scenario``: from attributes of the `scenario`, `model_name`,
         and/or `scenario_name` argument(s).
       - ``variable``: from the :attr:`~.Quantity.name` of `quantity`, if any.
       - ``unit``: from the :attr:`~.Quantity.units` of `quantity`, if any.

    3. The actions specified by the optional arguments `rename`, `collapse`, `replace`,
       `drop`, and `unit`, if any, are applied in that order.
    4. The resulting data frame is converted to :class:`pyam.IamDataFrame`.

    Parameters
    ----------
    scenario :
        Any object with :py:`model` and :py:`scenario` attributes of type :class:`str`,
        for instance an :class:`ixmp.Scenario` or
        :class:`~message_ix_models.util.scenarioinfo.ScenarioInfo`; **or** a
        :class:`str`, which is equivalent to `scenario_name`.
    quantity : .Quantity
        Quantity to convert to IAMC data structure.
    rename : dict, optional
        Mapping from dimension names in `quantity` (:class:`str`) to column names
        (:class:`str`); either IAMC dimension names, or others that are consumed by
        `collapse`.
    collapse : callable, optional
        Function that takes a :class:`pandas.DataFrame` and returns the same type.
        This function **may** collapse 2 or more dimensions, for example to construct
        labels for the IAMC ``variable`` dimension, or any other.
    replace : optional
        Values to be replaced and their replaced. Passed directly to
        :meth:`pandas.DataFrame.replace`.
    drop : str or collection of str, optional
        Columns to drop. Passed to :func:`.util.drop`, so if not given, all non-IAMC
        columns are dropped.
    unit : str, optional
        Label for the IAMC ``unit`` dimension. Passed to
        :func:`~.pyam.util.clean_units`.
    prepend_name : bool, optional
        If :any:`True`, the :attr:`.Quantity.name` of `quantity` is prepended to the
        IAMC ``variable`` dimension.
    model_name : str, optional
        Value for the IAMC ``model`` dimension.
    scenario_name : str, optional
        Value for the IAMC ``scenario`` dimension.

    Raises
    ------
    ValueError
        If the resulting data frame has duplicate keys in the IAMC dimensions.
        :class:`pyam.IamDataFrame` cannot handle such data.
    TypeError
        If both `scenario` and `scenario_name` are non-empty :class:`str`.
    """
    import pyam

    # Values to assign on all rows
    assign = dict(unit=quantity.units)
    if prepend_name:
        assign.update(variable=quantity.name)
    try:
        assign.update(model=scenario.model, scenario=scenario.scenario)
    except AttributeError:
        if scenario and scenario_name:
            raise TypeError(f"Both {scenario=!r} and {scenario_name=!r} given")
        assign.update(model=model_name or "", scenario=scenario or scenario_name or "")

    # - Convert to pd.DataFrame
    # - Rename one dimension to 'year' or 'time'
    # - Fill variable, unit, model, and scenario columns
    # - Replace values
    # - Apply the collapse callback, if given
    # - Drop any unwanted columns
    # - Clean units
    df = (
        quantity.to_series()
        .rename("value")
        .reset_index()
        .assign(**assign)
        .rename(columns=rename or dict())
        .pipe(collapse or util.collapse)
        .replace(replace, regex=True)
        .pipe(util.drop, columns=drop)
        .pipe(util.clean_units, unit)
    )

    # Raise exception for non-unique data
    duplicates = df.duplicated(subset=set(df.columns) - {"value"})
    if duplicates.any():
        raise ValueError(
            "Duplicate IAMC indices cannot be converted:\n"
            + str(df[duplicates].drop(columns=["model", "scenario"]))
        )

    return pyam.IamDataFrame(df)


@as_pyam.helper
def add_as_pyam(
    func,
    c: "Computer",
    quantities: Union[KeyLike, Iterable[KeyLike]],
    tag="iamc",
    /,
    **kwargs,
):
    """:meth:`.Computer.add` helper for :func:`.as_pyam`.

    Add conversion of one or more `quantities` to the IAMC data structure.

    Parameters
    ----------
    quantities : str or .Key or list of str or .Key
        Keys for quantities to transform.
    tag : str, optional
        Tag to append to new Keys.

    Other parameters
    ----------------
    kwargs :
        Any keyword arguments accepted by :func:`.as_pyam`.

    Returns
    -------
    list of .Key
        Each task converts a :class:`.Quantity` into a :class:`pyam.IamDataFrame`.
    """
    # Handle single vs. iterable of inputs
    if isinstance(quantities, (str, Key)):
        quantities = [quantities]
        multi_arg = False
    else:
        multi_arg = True

    if len(kwargs.get("replace", {})) and not isinstance(
        next(iter(kwargs["replace"].values())), dict
    ):
        kwargs["replace"] = dict(variable=kwargs.pop("replace"))
        warn(
            f"replace must be nested dict(), e.g. {repr(kwargs['replace'])}",
            DeprecationWarning,
        )

    # Check keys
    quantities = c.check_keys(*quantities)

    # The callable for the task. If pyam is not available, require_compat() above will
    # fail; so this will never be None
    comp = partial(func, **kwargs)

    keys = []
    for qty in quantities:
        # Key for the input quantity, e.g. foo:x-y-z
        key = Key(qty)

        # Key for the task/output, e.g. foo::iamc
        keys.append(Key(key.name, tag=tag))

        # Add the task and store the key
        c.add_single(keys[-1], (comp, "scenario", key))

    return tuple(keys) if multi_arg else keys[0]


@genno.operator.concat.register
def _(*args: pyam.IamDataFrame, **kwargs) -> "pyam.IamDataFrame":
    """Concatenate `args`, which must all be :class:`pyam.IamDataFrame`.

    Otherwise, equivalent to :func:`genno.operator.concat`.
    """
    # Use pyam.concat() top-level function
    return pyam.concat(args, **kwargs)


def quantity_from_iamc(
    qty: Union["AnyQuantity", "pyam.IamDataFrame", "pandas.DataFrame"],
    variable: str,
    *,
    fail: Union[int, str] = "warning",
) -> "AnyQuantity":
    """Extract data for a single measure from `qty` with IAMC-like structure.

    Parameters
    ----------
    qty :
        Must have at least 2 dimensions named ‘v’ (or ‘variable’, any case) and ‘u’
        (or ‘unit’, any case).
    variable : str
        Regular expression to match full labels on the ``v`` dimension of `qty`. If the
        expression contains match groups, they are used to rewrite ``v`` labels: only
        the contents of the first match group are kept. This may be used to discard a
        portion of the label.

    Returns
    -------
    .Quantity
        The ‘variable’ dimension contains reduced labels.
        The :attr:`.Quantity.units` attribute contains the unique units for the subset
        of data.

    See also
    --------
    unique_units_from_dim
    """
    import pandas as pd

    from genno.operator import relabel, select, unique_units_from_dim

    from .util import IAMC_DIMS

    if isinstance(qty, pd.DataFrame):
        # Convert pandas.DataFrame to pyam.IamDataFrame
        qty = pyam.IamDataFrame(qty)
    if isinstance(qty, pyam.IamDataFrame):
        # Convert IamDataFrame to Quantity
        df = qty.as_pandas()
        qty = genno.Quantity(df.set_index(list(IAMC_DIMS & set(df.columns)))["value"])

    # Identify a dimension whose name is in `targets`
    def identify_dim(targets: Collection[str]) -> str:
        result = list(filter(lambda d: d.lower() in targets, qty.dims))
        if len(result) != 1:
            raise ValueError(
                f"cannot identify 1 unique dimension for {targets!r} among "
                f"{qty.dims!r}; found {result!r}"
            )
        return result[0]

    v_dim = identify_dim(("v", "variable"))
    u_dim = identify_dim(("u", "unit"))

    # Compile expression
    expr = re.compile(variable)
    has_group = expr.groups > 0

    # Process each label along v_dim
    variables, replacements = [], {}
    for var in qty.coords[v_dim].data:
        if match := expr.fullmatch(var):
            variables.append(match.group(0))
            replacements.update({match.group(0): match.group(1)} if has_group else {})

    if not variables:
        log.warning(
            f"0 of {len(qty.coords[v_dim])} labels on dimension {v_dim!r} were a full "
            f"match for {expr!r}"
        )

    return (
        qty.pipe(select, {v_dim: variables})
        .pipe(relabel, {v_dim: replacements})
        .pipe(unique_units_from_dim, u_dim, fail=fail)
    )


@genno.operator.write_report.register
def _(quantity: pyam.IamDataFrame, path, kwargs=None) -> None:
    """Write  `obj` to the file at `path`.

    If `obj` is a :class:`pyam.IamDataFrame` and `path` ends with ".csv" or ".xlsx",
    use :mod:`pyam` methods to write the file to CSV or Excel format, respectively.
    Otherwise, equivalent to :func:`genno.operator.write_report`.
    """
    path = Path(path)

    if kwargs is not None and len(kwargs):
        raise NotImplementedError(
            "Keyword arguments to write_report(pyam.IamDataFrame, …)"
        )

    if path.suffix == ".csv":
        quantity.to_csv(path)
    elif path.suffix == ".xlsx":
        quantity.to_excel(path)
    else:
        raise ValueError(
            f"pyam.IamDataFrame can be written to .csv or .xlsx, not {path.suffix}"
        )


def __getattr__(name: str):
    if name in ("concat", "write_report"):
        warn(
            f"Importing {name!r} from genno.compat.pyam.operator; import from "
            "genno.operator instead.",
            DeprecationWarning,
            2,
        )

        return getattr(genno.operator, name)
    else:
        raise AttributeError(name)



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Hashable, Optional, Sequence
from warnings import warn

import plotnine as p9

import genno
from genno.compat.pandas import disable_copy_on_write
from genno.core.computer import Computer
from genno.core.key import KeyLike

log = logging.getLogger(__name__)


class Plot(ABC):
    """Class for plotting using :mod:`plotnine`."""

    #: File name base for saving the plot.
    basename = ""
    #: File extension; determines file format.
    suffix = ".pdf"
    #: Path for file output. If it is not set, :meth:`save` will populate it with a
    #: value constructed from :py:`config["output_dir"]`, :attr:`basename`, and
    #: :attr:`suffix`. The implementation of :meth:`generate` in a Plot sub-class may
    #: assign any other value, for instance one constructed at runtime from the
    #: :attr:`inputs`.
    path: Optional[Path] = None
    #: :class:`Keys <.Key>` referring to :class:`Quantities <.Quantity>` or other inputs
    #: accepted by :meth:`generate`.
    inputs: Sequence[Hashable] = []
    #: Keyword arguments for :any:`plotnine.ggplot.save`.
    save_args: dict[str, Any] = dict(verbose=False)

    # TODO add static geoms automatically in generate()
    __static: Sequence = []

    def save(self, config, *args, **kwargs) -> Optional[Path]:
        """Prepare data, call :meth:`.generate`, and save to file.

        This method is used as the callable in the task generated by :meth:`.add_tasks`.

        .. versionadded:: 1.24.1

           This method uses :func:`.disable_copy_on_write` to work around
           `has2k1/mizani#38 <https://github.com/has2k1/mizani/issues/38>`_. This may
           cause issues if other computations (for instance, of the inputs to the Plot)
           rely on Pandas' copy-on-write behaviour being enabled.
        """
        self.path = self.path or (
            config["output_dir"] / f"{self.basename}{self.suffix}"
        )

        missing = tuple(filter(lambda arg: isinstance(arg, str), args))
        if len(missing):
            log.error(
                f"Missing input(s) {missing!r} to plot {self.basename!r}; no output"
            )
            return None

        # Convert Quantity arguments to pd.DataFrame for use with plotnine
        _args = map(
            lambda arg: arg
            if not isinstance(arg, genno.Quantity)
            else arg.to_series()
            .rename(arg.name or "value")
            .reset_index()
            .assign(unit=f"{arg.units:~}"),
            args,
        )

        plot_or_plots = self.generate(*_args, **kwargs)

        if not plot_or_plots:
            log.info(
                f"{self.__class__.__name__}.generate() returned {plot_or_plots!r}; no "
                "output"
            )
            return None

        log.info(f"Save to {self.path}")

        with disable_copy_on_write(f"{__name__}.Plot.save()"):
            try:
                # Single plot
                plot_or_plots.save(self.path, **self.save_args)
            except AttributeError:
                # Iterator containing 0 or more plots
                p9.save_as_pdf_pages(plot_or_plots, self.path, **self.save_args)

        return self.path

    @classmethod
    def make_task(cls, *inputs):
        """Return a task :class:`tuple` to add to a Computer.

        .. deprecated:: 1.18.0

           Use :func:`add_tasks` instead.

        Parameters
        ----------
        *inputs : `.Key` or str or hashable, optional
            If provided, overrides the :attr:`inputs` property of the class.

        Returns
        -------
        tuple
            - The first, callable element of the task is :meth:`save`.
            - The second element is ``"config"``, to access the configuration of the
              Computer.
            - The third and following elements are the `inputs`.
        """
        inputs_repr = ",".join(map(repr, inputs))
        warn(
            f"Plot.make_task(…). Use: Computer.add(…, {cls.__name__}"
            + (", " if inputs_repr else "")
            + f"{inputs_repr})",
            DeprecationWarning,
        )
        return tuple([cls().save, "config"] + (list(inputs) if inputs else cls.inputs))

    @classmethod
    def add_tasks(
        cls, c: Computer, key: KeyLike, *inputs, strict: bool = False
    ) -> KeyLike:
        """Add a task to `c` to generate and save the Plot.

        Analogous to :meth:`.Operator.add_tasks`.
        """
        _inputs = list(inputs if inputs else cls.inputs)
        if strict:
            _inputs = c.check_keys(*_inputs)
        return c.add_single(key, cls().save, "config", *_inputs)

    @abstractmethod
    def generate(self, *args, **kwargs):
        """Generate and return the plot.

        A subclass of Plot **must** implement this method.

        Parameters
        ----------
        args : sequence of pandas.DataFrame or other
            One argument is given corresponding to each of the :attr:`inputs`.

            Because :mod:`plotnine` operates on pandas data structures, :meth:`save`
            automatically converts any :class:`.Quantity` inputs to
            :class:`pandas.DataFrame` before they are passed to :meth:`generate`.
        """



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information ---------------------------------------------------------------

project = "genno"
copyright = "2018–2024, Genno contributors"
author = "Genno contributors"


# -- General configuration -------------------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions coming
# with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    # First-party
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    # Others
    "genno.compat.sphinx.autodoc_operator",
    "genno.compat.sphinx.rewrite_refs",
    "IPython.sphinxext.ipython_directive",
]

# List of patterns, relative to source directory, that match files and directories to
# ignore when looking for source files. This pattern also affects html_static_path and
# html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

nitpicky = True

rst_prolog = """
.. role:: py(code)
   :language: python
"""

# Paths that contain templates, relative to the current directory.
templates_path = ["_templates"]

# -- Options for HTML output -----------------------------------------------------------

html_css_files = ["custom.css"]

html_logo = "_static/hammer-duotone.svg"

html_static_path = ["_static"]

html_theme = "furo"

GH_URL = "https://github.com/khaeru/genno"
html_theme_options = dict(
    footer_icons=[
        {
            "name": "GitHub",
            "url": GH_URL,
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
    top_of_page_buttons=["view", "edit"],
    source_repository=GH_URL,
    source_branch="main",
    source_directory="doc/",
)

html_title = "genno"

# -- Options for genno.compat.sphinx.reference_alias -----------------------------------

# Mapping from expression → replacement.
# Order matters here; earlier entries are matched first.
reference_aliases = {
    r"Quantity\.units": "genno.core.base.UnitsMixIn.units",
    "Quantity": "genno.core.attrseries.AttrSeries",
    "AnyQuantity": ":data:`genno.core.quantity.AnyQuantity`",
    #
    # Many projects (including Sphinx itself!) do not have a py:module target in for the
    # top-level module in objects.inv. Resolve these using :doc:`index` or similar for
    # each project.
    "dask$": ":std:doc:`dask:index`",
    "pint$": ":std:doc:`pint <pint:index>`",
    "plotnine$": ":class:`plotnine.ggplot`",
    "pyarrow$": ":std:doc:`pyarrow <pyarrow:python/index>`",
    "pyam$": ":std:doc:`pyam:index`",
    "sphinx$": ":std:doc:`sphinx <sphinx:index>`",
}

# -- Options for sphinx.ext.extlinks ---------------------------------------------------

extlinks = {
    "issue": ("https://github.com/khaeru/genno/issues/%s", "#%s"),
    "pull": ("https://github.com/khaeru/genno/pull/%s", "PR #%s"),
    "gh-user": ("https://github.com/%s", "@%s"),
}
extlinks_detect_hardcoded_links = False

# -- Options for sphinx.ext.intersphinx ------------------------------------------------

intersphinx_mapping = {
    "dask": ("https://docs.dask.org/en/stable", None),
    "ixmp": ("https://docs.messageix.org/projects/ixmp/en/latest", None),
    "joblib": ("https://joblib.readthedocs.io/en/latest", None),
    "graphviz": ("https://graphviz.readthedocs.io/en/stable", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "message_ix": ("https://docs.messageix.org/en/latest", None),
    "message-ix-models": ("https://docs.messageix.org/projects/models/en/latest", None),
    "nbclient": ("https://nbclient.readthedocs.io/en/latest", None),
    "nbformat": ("https://nbformat.readthedocs.io/en/latest", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
    "pint": ("https://pint.readthedocs.io/en/stable", None),
    "platformdirs": ("https://platformdirs.readthedocs.io/en/latest", None),
    "plotnine": ("https://plotnine.org", None),
    "pyam": ("https://pyam-iamc.readthedocs.io/en/stable", None),
    "pyarrow": ("https://arrow.apache.org/docs", None),
    "python": ("https://docs.python.org/3", None),
    "pytest": ("https://docs.pytest.org/en/stable", None),
    "sdmx1": ("https://sdmx1.readthedocs.io/en/stable", None),
    "sparse": ("https://sparse.pydata.org/en/stable", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "xarray": ("https://docs.xarray.dev/en/stable", None),
}

# -- Options for sphinx.ext.napoleon ---------------------------------------------------

napoleon_preprocess_types = True
napoleon_type_aliases = {
    # Standard library
    "callable": "typing.Callable",
    "collection": "collections.abc.Collection",
    "hashable": "collections.abc.Hashable",
    "iterable": "collections.abc.Iterable",
    "mapping": "collections.abc.Mapping",
    "sequence": "collections.abc.Sequence",
    "Path": "pathlib.Path",
    # This package
    "KeyLike": "genno.core.key.KeyLike",
    # Others
    "Code": "sdmx.model.common.Code",
}

# -- Options for sphinx.ext.todo -------------------------------------------------------

todo_include_todos = True



import logging
import re
from functools import partial, singledispatch
from itertools import chain, compress
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Callable,
    Generator,
    Hashable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    SupportsInt,
    Union,
)
from warnings import warn

from .attrseries import AttrSeries
from .sparsedataarray import SparseDataArray

if TYPE_CHECKING:
    from .quantity import AnyQuantity


log = logging.getLogger(__name__)

#: Regular expression for valid key strings.
EXPR = re.compile(r"^(?P<name>[^:]+)(:(?P<dims>([^:-]*-)*[^:-]+)?(:(?P<tag>[^:]*))?)?$")

#: Regular expression for non-keylike strings.
BARE_STR = re.compile(r"^\s*(?P<name>[^:]+)\s*$")


@singledispatch
def _name_dims_tag(value) -> tuple[str, tuple[str, ...], Optional[str]]:
    """Convert various `value`s into (name, dims, tag) tuples.

    Helper for :meth:`.Key.__init__`.
    """
    raise TypeError(type(value))


@_name_dims_tag.register
def _(value: str):
    """Parse a string that may contain a Key expression."""
    match = EXPR.match(value)
    if match is None:
        raise ValueError(f"Invalid key expression: {repr(value)}")
    groups = match.groupdict()
    return (
        groups["name"],
        tuple() if not groups["dims"] else tuple(groups["dims"].split("-")),
        groups["tag"],
    )


@_name_dims_tag.register(AttrSeries)
@_name_dims_tag.register(SparseDataArray)
def _(value: "AnyQuantity"):  # register() only handles bare AnyQuantity in Python ≥3.11
    """Return (name, dims, tag) that describe an existing Quantity."""
    return str(value.name), tuple(map(str, value.dims)), None


class Key:
    """A hashable key for a quantity that includes its dimensionality."""

    _name: str
    _dims: tuple[str, ...]
    _tag: Optional[str]

    def __init__(
        self,
        name_or_value: Union[str, "Key", "AnyQuantity"],
        dims: Iterable[str] = [],
        tag: Optional[str] = None,
        _fast: bool = False,
    ):
        if _fast:
            # Fast path: don't handle arguments
            assert isinstance(name_or_value, str)
            self._name = name_or_value
            self._dims = tuple(dims)
            self._tag = tag or None
        else:
            # Convert various values into a (name, dims, tags)
            self._name, _dims, _tag = _name_dims_tag(name_or_value)

            # Check for conflicts between dims inferred from name_or_value and any
            # direct argument
            # TODO handle resolveable combinations without raising exceptions
            if bool(_dims) and bool(dims):
                raise ValueError(
                    f"Conflict: {dims = } argument vs. {_dims!r} from {name_or_value!r}"
                )
            elif bool(_tag) and bool(tag):
                raise ValueError(
                    f"Conflict: {tag = } argument vs. {_tag!r} from {name_or_value!r}"
                )

            self._dims = _dims or tuple(dims)
            self._tag = _tag or tag

        # Pre-compute string representation and hash
        self._str = (
            self._name
            + ":"
            + "-".join(self._dims)
            + (f":{self._tag}" if self._tag else "")
        )
        self._hash = hash(self._str)

    # Class methods

    @classmethod
    def bare_name(cls, value) -> Optional[str]:
        """If `value` is a bare name (no dims or tags), return it; else :obj:`None`."""
        if not isinstance(value, str):
            return None
        match = BARE_STR.match(value)
        return match.group("name") if match else None

    @classmethod
    def from_str_or_key(
        cls,
        value: Union[str, "Key", "AnyQuantity"],
        drop: Union[Iterable[str], bool] = [],
        append: Iterable[str] = [],
        tag: Optional[str] = None,
    ) -> "Key":
        """Return a new Key from *value*.

        .. versionchanged:: 1.18.0

           Calling :meth:`from_str_or_key` with a single argument is no longer
           necessary; simply give the same `value` as an argument to :class:`Key`.

           The class method is retained for convenience when calling with multiple
           arguments. However, the following are equivalent and may be more readable:

           .. code-block:: python

              k1 = Key("foo:a-b-c:t1", drop="b", append="d", tag="t2")
              k2 = Key("foo:a-b-c:t1").drop("b").append("d)"

        Parameters
        ----------
        value : str or .Key
            Value to use to generate a new Key.
        drop : list of str or :obj:`True`, optional
            Existing dimensions of *value* to drop. See :meth:`drop`.
        append : list of str, optional
            New dimensions to append to the returned Key. See :meth:`append`.
        tag : str, optional
            Tag for returned Key. If *value* has a tag, the two are joined
            using a '+' character. See :meth:`add_tag`.

        Returns
        -------
        :class:`Key`
        """
        base = cls(value)

        # Return quickly if no further manipulations are required
        if not any([drop, append, tag]):
            warn(
                "Calling Key.from_str_or_key(value) with no other arguments is no "
                "longer necessary; simply use Key(value)",
                FutureWarning,
                stacklevel=2,
            )
            return base

        # mypy is fussy here
        drop_args: tuple[Union[str, bool], ...] = tuple(
            [drop] if isinstance(drop, bool) else drop
        )

        # Drop and append dimensions; add tag
        return base.drop(*drop_args).append(*tuple(append)).add_tag(tag)

    @classmethod
    def product(cls, new_name: str, *keys, tag: Optional[str] = None) -> "Key":
        """Return a new Key that has the union of dimensions on *keys*.

        Dimensions are ordered by their first appearance:

        1. First, the dimensions of the first of the *keys*.
        2. Next, any additional dimensions in the second of the *keys* that
           were not already added in step 1.
        3. etc.

        Parameters
        ----------
        new_name : str
            Name for the new Key. The names of *keys* are discarded.
        """
        # Iterable of dimension names from all keys, in order, with repetitions
        dims = chain(*map(lambda k: cls(k).dims, keys))

        # Return new key. Use dict to keep only unique *dims*, in same order
        return cls(new_name, dict.fromkeys(dims).keys()).add_tag(tag)

    def __add__(self, other: str) -> "Key":
        if not isinstance(other, str):
            raise TypeError(type(other))
        return self.add_tag(other)

    def __sub__(self, other: Union[str, Iterable[str]]) -> "Key":
        return self.remove_tag(*((other,) if isinstance(other, str) else other))

    def __mul__(self, other: Union[str, "Key", Sequence[str]]) -> "Key":
        if isinstance(other, str):
            other_dims: Sequence[str] = (other,)
        elif isinstance(other, Key):
            other_dims = other.dims
        elif isinstance(other, Sequence):
            other_dims = other
        else:
            raise TypeError(type(other))

        return self.append(*other_dims)

    def __truediv__(self, other: Union[str, "Key", Sequence[str]]) -> "Key":
        if isinstance(other, str):
            other_dims: Sequence[str] = (other,)
        elif isinstance(other, Key):
            other_dims = other.dims
        elif isinstance(other, Sequence):
            other_dims = other
        else:
            raise TypeError(type(other))

        return self.drop(*other_dims)

    def __repr__(self) -> str:
        """Representation of the Key, e.g. '<name:dim1-dim2-dim3:tag>."""
        return f"<{self._str}>"

    def __str__(self) -> str:
        """String equivalent of the Key, e.g. 'name:dim1-dim2-dim3:tag'."""
        return self._str  # Return the pre-computed value

    def __hash__(self):
        """Key hashes the same as :py:`str(Key)`."""
        return self._hash

    def __eq__(self, other) -> bool:
        """Key is equal to :py:`str(Key)`."""
        try:
            other = Key(other)
        except TypeError:
            return NotImplemented

        return (
            (self.name == other.name)
            and (set(self.dims) == set(other.dims))
            and (self.tag == other.tag)
        )

    # Less-than and greater-than operations, for sorting
    def __lt__(self, other) -> bool:
        if isinstance(other, Key):
            return str(self.sorted) < str(other.sorted)
        elif isinstance(other, str):
            return str(self.sorted) < other
        else:
            return NotImplemented

    def __gt__(self, other) -> bool:
        if isinstance(other, Key):
            return str(self.sorted) > str(other.sorted)
        elif isinstance(other, str):
            return str(self.sorted) > other
        else:
            return NotImplemented

    @property
    def name(self) -> str:
        """Name of the quantity, :class:`str`."""
        return self._name

    @property
    def dims(self) -> tuple[str, ...]:
        """Dimensions of the quantity, :class:`tuple` of :class:`str`."""
        return self._dims

    @property
    def tag(self) -> Optional[str]:
        """Quantity tag, :class:`str` or :obj:`None`."""
        return self._tag

    @property
    def sorted(self) -> "Key":
        """A version of the Key with its :attr:`.dims` :func:`sorted`."""
        return Key(self._name, sorted(self._dims), self._tag, _fast=True)

    def rename(self, name: str) -> "Key":
        """Return a Key with a replaced `name`."""
        return Key(name, self._dims, self._tag, _fast=True)

    def drop(self, *dims: Union[str, bool]) -> "Key":
        """Return a new Key with `dims` dropped."""
        return Key(
            self._name,
            [] if dims == (True,) else filter(lambda d: d not in dims, self._dims),
            self._tag,
            _fast=True,
        )

    def drop_all(self) -> "Key":
        """Return a new Key with all dimensions dropped / zero dimensions."""
        return Key(self._name, tuple(), self._tag, _fast=True)

    def append(self, *dims: str) -> "Key":
        """Return a new Key with additional dimensions `dims`."""
        return Key(self._name, list(self._dims) + list(dims), self._tag, _fast=True)

    def add_tag(self, tag: Optional[str]) -> "Key":
        """Return a new Key with `tag` appended."""
        return Key(
            self._name, self._dims, "+".join(filter(None, [self._tag, tag])), _fast=True
        )

    def iter_sums(self) -> Generator[tuple["Key", Callable, "Key"], None, None]:
        """Generate (key, task) for all possible partial sums of the Key."""
        from genno.operator import sum

        for agg_dims, others in combo_partition(self.dims):
            yield (
                Key(self._name, agg_dims, self.tag, _fast=True),
                partial(sum, dimensions=others, weights=None),
                self,
            )

    def remove_tag(self, *tags: str) -> "Key":
        """Return a key with any of `tags` dropped.

        Raises
        ------
        ValueError
            If none of `tags` are in :attr:`.tags`.
        """
        new_tags = tuple(filter(lambda t: t not in tags, (self.tag or "").split("+")))
        new_tag = "+".join(new_tags) if new_tags else None
        if new_tag == self.tag:
            raise ValueError(f"No existing tags {tags!r} to remove")
        return Key(self._name, self._dims, new_tag, _fast=True)


@_name_dims_tag.register
def _(value: Key):
    """Return the (name, dims, tag) of an existing Key."""
    return value._name, value._dims, value._tag


class KeySeq:
    """Utility class for generating similar :class:`Keys <.Key>`."""

    #: Base :class:`.Key` of the sequence.
    base: Key

    # Keys that have been created.
    _keys: dict[Hashable, Key]

    def __init__(self, *args, **kwargs):
        self.base = Key(*args, **kwargs)
        self._keys = {}

    def _next_int_tag(self) -> int:
        return max([-1] + [t for t in self._keys if isinstance(t, int)]) + 1

    def __next__(self) -> Key:
        return self[self._next_int_tag()]

    def __call__(self, value: Optional[Hashable] = None) -> Key:
        return next(self) if value is None else self[value]

    def __getitem__(self, value: Hashable) -> Key:
        tag = int(value) if isinstance(value, SupportsInt) else str(value)
        result = self._keys[tag] = self.base + str(tag)
        return result

    def __repr__(self) -> str:
        return f"<KeySeq from '{self.base!s}'>"

    @property
    def keys(self) -> MappingProxyType:
        """Read-only view of previously-created :class:`Keys <.Key>`.

        In the form of a :class:`dict` mapping tags (:class:`int` or :class:`str`) to
        :class:`.Key` values.
        """
        return MappingProxyType(self._keys)

    @property
    def prev(self) -> Key:
        """The most recently created :class:`.Key`."""
        return next(reversed(self._keys.values()))

    # Access to Key properties
    @property
    def name(self) -> str:
        """Name of the :attr:`.base` Key."""
        return self.base.name

    @property
    def dims(self) -> tuple[str, ...]:
        """Dimensions of the :attr:`.base` Key."""
        return self.base.dims

    @property
    def tag(self) -> Optional[str]:
        """Tag of the :attr:`.base` Key."""
        return self.base.tag

    def __add__(self, other: str) -> "KeySeq":
        return KeySeq(self.base + other)

    def __mul__(self, other) -> "KeySeq":
        return KeySeq(self.base * other)

    def __sub__(self, other: Union[str, Iterable[str]]) -> "KeySeq":
        return KeySeq(self.base - other)

    def __truediv__(self, other) -> "KeySeq":
        return KeySeq(self.base / other)


#: Type shorthand for :class:`Key` or any other value that can be used as a key.
KeyLike = Union[Key, str]


def combo_partition(iterable):
    """Yield pairs of lists with all possible subsets of *iterable*."""
    # Format string for binary conversion, e.g. '04b'
    fmt = "0{}b".format(len(iterable))
    for n in range(2 ** len(iterable) - 1):
        # Two binary lists
        a, b = zip(*[(v, not v) for v in map(int, format(n, fmt))])
        yield list(compress(iterable, a)), list(compress(iterable, b))


def iter_keys(value: Union[KeyLike, tuple[KeyLike, ...]]) -> Iterator[Key]:
    """Yield :class:`Keys <Key>` from `value`.

    Raises
    ------
    TypeError
        `value` is not an iterable of :class:`Key`.

    See also
    --------
    .Computer.add
    """
    if isinstance(value, (Key, str)):
        yield Key(value)
        tmp: Iterator[KeyLike] = iter(())
    else:
        tmp = iter(value)
    for element in tmp:
        if not isinstance(element, Key):
            raise TypeError(type(element))
        yield element


def single_key(value: Union[KeyLike, tuple[KeyLike, ...], Iterator]) -> Key:
    """Ensure `value` is a single :class:`Key`.

    Raises
    ------
    TypeError
        `value` is not a :class:`Key` or 1-tuple of :class:`Key`.

    See also
    --------
    .Computer.add
    """
    if isinstance(value, (Key, str)):
        return Key(value)

    tmp = iter(value)
    try:
        result = next(tmp)
    except StopIteration:
        raise TypeError("Empty iterable")
    else:
        try:
            next(tmp)
        except StopIteration:
            pass
        else:
            raise TypeError("Iterable of length >1")

    if isinstance(result, Key):
        return result
    else:
        raise TypeError(type(result))
